#
# This file was generated using xslt from its XML file
#
# Copyright 2009, Associated Universities Inc., Washington DC
#
import sys
import os
from  casac import *
import string
from taskinit import casalog
#from taskmanager import tm
import task_msmoments
def msmoments(infile='', moments=[0], antenna='', field='', spw='', includemask=-1, excludemask=-1, outfile='', overwrite=False):

        """Compute moments from an MS
        The spectral moment distributions at each row in input MS are
        determined. Input MS must have FLOAT_DATA column, i.e. 
        autocorrelation data.  
        See the cookbook and User Reference Manual for
        mathematical details.

        The main control of the calculation is given by parameter
        moments:
	
        moments=-1  - mean value of the spectrum
        moments=0   - integrated value of the spectrum
        moments=1   - intensity weighted coordinate;traditionally used to get 
                      'velocity fields'
        moments=2   - intensity weighted dispersion of the coordinate; traditionally
                      used to get "velocity dispersion"
        moments=3   - median of I
        moments=4   - median coordinate
        moments=5   - standard deviation about the mean of the spectrum
        moments=6   - root mean square of the spectrum
        moments=7   - absolute mean deviation of the spectrum
        moments=8   - maximum value of the spectrum
        moments=9   - coordinate of the maximum value of the spectrum
        moments=10  - minimum value of the spectrum
        moments=11  - coordinate of the minimum value of the spectrum

        Note that includemask and excludemask cannot set simultaneously. 

        Keyword arguments:
        infile -- Name of input MS data
                default: none; example: infile="OrionS_rawACSmod"
        moments -- List of moments you would like to compute
                default: 0 (integrated spectrum);example: moments=[0,1]
                see list above
        antenna -- antenna name or id that the user wants to compute moments
                default: '' (all antennae)
        field -- field name or id that the user wants to compute moments
                default: '' (all fields)
        spw -- spectral window id that the user wants to compute moments
                default: '' (all spectral windows)

        includemask -- List of masks to include
                default: [-1] (include all channels); example=[2,100]
        excludemask -- List of masks to exclude
                default: [-1] (don't exclude channels); example=[100,200]
        outfile -- Output MS file name (or root for multiple moments)
                default: '' (input+auto-determined suffix);example: outfile='source_moment'
        overwrite -- Overwrite existing output files
                default: false

        Example for finding the 1-momment, intensity-weighted
        coordinate, often used for finding velocity fields.
        msmoments( infile='mydata', moment=1, outfile='velocityfields' )


        """
        if type(moments)==int: moments=[moments]

#
#    The following is work around to avoid a bug with current python translation
#
        mytmp = {}

        mytmp['infile'] = infile
        mytmp['moments'] = moments
        mytmp['antenna'] = antenna
        mytmp['field'] = field
        mytmp['spw'] = spw
        mytmp['includemask'] = includemask
        mytmp['excludemask'] = excludemask
        mytmp['outfile'] = outfile
        mytmp['overwrite'] = overwrite
	pathname='file:///'+os.environ.get('CASAPATH').split()[0]+'/xml/'
        trec = casac.utils().torecord(pathname+'msmoments.xml')

        casalog.origin('msmoments')
        if trec.has_key('msmoments') and casac.utils().verify(mytmp, trec['msmoments']) :
	    result = task_msmoments.msmoments(infile, moments, antenna, field, spw, includemask, excludemask, outfile, overwrite)

	else :
	  result = False
        return result
