#
# This file was generated using xslt from its XML file
#
# Copyright 2009, Associated Universities Inc., Washington DC
#
import sys
import os
from  casac import *
import string
from taskinit import casalog
#from taskmanager import tm
import task_sdtpimaging
def sdtpimaging(infile='', calmode='none', masklist=[], blpoly=1, backup=True, flaglist=[], antenna='', spw=0, stokes='', createimage=False, outfile='', imsize=[256, 256], cell=['1.0arcmin', '1.0arcmin'], phasecenter='', ephemsrcname='', pointingcolumn='direction', gridfunction='BOX', plotlevel=0):

        """SD task: do a simple calibration (baseline subtraction) and imaging for total power data 
Keyword arguments:
infile -- name of an input SD Measurementset
        example: 'm100.tp.ms'
calmode -- SD calibration mode (currently only baseline subtraction)
        options: 'baseline','none'
        default: 'none'
        example: choose mode 'none' if you have
                 already calibrated and want to do
                 plotting nd/or imaging 
    >>> calmode='baseline' expandable parameters
        masklist -- numbers of integrations from each edge of each scan
                    to be included for baseline fitting
                default: [] (no edge. should define positive number)
                example: [30,30] or [30] 
                         used first 30 rows and last 30 rows of each scan
                         for the baseline 
        blpoly -- polynomial order of the baseline fit
                default: (int) 1
                example: any number >=0
        backup -- set True to create backup of input data
                options: (bool) True, False
                default: True
flaglist -- list of scan numbers to flag (ranges can be accepted)  
        default: [] (use all scans)
        example: [[0,3],80]
                 flag the scan range [0,3] = [0,1,2,3] and scan 80 
antenna -- select data based on antenna names or IDs
        default: '' (use all antennas)
        example: antenna='0,1' (antenna ID 0 and 1)
                 antenna='DV01'
        WARNING: currently baseline subtraction properly 
                 only one of the antennas.
spw -- spectral window ID for imaging (should have only one channel)
        default: 0
        example: spw=11 (SPW ID 11)
stokes -- stokes parameters or polarization types to image
        default: '' (Stokes I)
        example: stokes='XX' (image plane of linear polarization, XX)
                 stokes='XXYY' (image cube with XX and YY image in each plane)
                 stokes='I' (Stokes I image = total intensity)
createimage -- do imaging? 
        options: (bool) True, False
        default: False 
    >>> createimage=True expandable parameters
        outfile -- name of output image
                default: ''
                example: 'mySDimage.im'
        imsize -- x and y image size in pixels, symmetric for single value
                default: [256,256]
                example: imsize=200 (equivalent to [200,200])
        cell -- x and y cell size. default unit arcmin
                default: '1.0arcmin'
                example: cell=['0.2arcmin, 0.2arcmin']
                         cell='0.2arcmin' (equivalent to example above)
        phasecenter -- image phase center: direction measure or field ID
                default: 0
                example: 'J2000 13h44m00 -17d02m00', 'AZEL -123d48m29 15d41m41'
        ephemsrcname -- ephemeris source name of moving source to use to
                        correct movements of source direction during
                        observation.
                default: ''
                         if the source name in the data matches one of the 
                         known solar objects by the system, this task 
                         automatically set the source name. 
                example: 'mars'
        pointingcolumn -- pointing data column to use
                options: 'direction', 'target', 'pointing_offset', 
                        'source_offset', 'encoder' 
                default: 'direction'
        gridfunction -- gridding function for imaging
                options: 'BOX' (Box-car), 'SF' (Spheroidal), 
                         'PB' (Primary-beam), 'GAUSS' (Gaussian),
                         'GJINC' (Gaussian*Jinc)
                default: 'BOX'
                example: 'SF'
plotlevel -- control for plotting of results
        options: (int) 0=none, 1=some, 2=more, <0=hardcopy
        default: 0 (no plotting)
        example: plotlevel<0 as abs(plotlevel), e.g.
                 -1: hardcopy plot 
                     (will be named <infile>_scans.eps)
                  1: plot raw data, calibrated data 
                     (for calmode='baseline)
                     plot raw or if exist calibrated data 
                     (for calmode='none')
                  2: plot raw data, progressively display baseline 
                     fitting for each scan, and final calibrated data 
                     (for calmode='baseline')  


-----------------
Gridding Kernel
-----------------
The parameter gridfunction sets gridding function (convolution kernel)
for imaging. Currently, the task supports 'BOX' (Box-car), 'SF' (Prolate
Spheroidal Wave Function), 'GAUSS' (Gaussian), 'GJINC' (Gaussian*Jinc),
where Jinc(x) = J_1(pi*x/c)/(pi*x/c) with a first order Bessel function
J_1, and 'PB' (Primary Beam). For 'PB', correct antenna informations
should be included in input file. 
Sub-parameters for convolution functions cannot be specified in this
task. To costomize your convolution function, please do imaging using
sdimaging task or imager tool.
  
        """
        if type(masklist)==int: masklist=[masklist]
        if type(flaglist)==int: flaglist=[flaglist]
        if type(imsize)==int: imsize=[imsize]
        if type(cell)==float: cell=[cell]

#
#    The following is work around to avoid a bug with current python translation
#
        mytmp = {}

        mytmp['infile'] = infile
        mytmp['calmode'] = calmode
        mytmp['masklist'] = masklist
        mytmp['blpoly'] = blpoly
        mytmp['backup'] = backup
        mytmp['flaglist'] = flaglist
        mytmp['antenna'] = antenna
        mytmp['spw'] = spw
        mytmp['stokes'] = stokes
        mytmp['createimage'] = createimage
        mytmp['outfile'] = outfile
        mytmp['imsize'] = imsize
        if type(cell) == str :
           mytmp['cell'] = casac.quanta().quantity(cell)
        else :
           mytmp['cell'] = cell
        mytmp['phasecenter'] = phasecenter
        mytmp['ephemsrcname'] = ephemsrcname
        mytmp['pointingcolumn'] = pointingcolumn
        mytmp['gridfunction'] = gridfunction
        mytmp['plotlevel'] = plotlevel
	pathname='file:///'+os.environ.get('CASAPATH').split()[0]+'/xml/'
        trec = casac.utils().torecord(pathname+'sdtpimaging.xml')

        casalog.origin('sdtpimaging')
        if trec.has_key('sdtpimaging') and casac.utils().verify(mytmp, trec['sdtpimaging']) :
	    result = task_sdtpimaging.sdtpimaging(infile, calmode, masklist, blpoly, backup, flaglist, antenna, spw, stokes, createimage, outfile, imsize, cell, phasecenter, ephemsrcname, pointingcolumn, gridfunction, plotlevel)

	else :
	  result = False
        return result
